from __future__ import annotations


class PlayStationNotFoundError(Exception):
    def __init__(self, message="Playstation not found on this network."):
        super().__init__(message)


class PlayStatonOnStandbyError(Exception):
    def __init__(self, playstation_ip):
        message = f"Playstation {'at '+playstation_ip+' ' if playstation_ip else ''}is on standby."
        super().__init__(message)



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class TelemetryPacket:
    position_x: float
    position_y: float
    position_z: float
    velocity_x: float
    velocity_y: float
    velocity_z: float
    rotation_x: float
    rotation_y: float
    rotation_z: float
    orientation: float
    ang_vel_x: float
    ang_vel_y: float
    ang_vel_z: float
    body_height: float
    engine_rpm: float
    iv: float
    fuel_level: float
    fuel_capacity: float
    speed_mps: float
    boost_pressure: float
    oil_pressure: float
    water_temp: float
    oil_temp: float
    tire_fl_temp: float
    tire_fr_temp: float
    tire_rl_temp: float
    tire_rr_temp: float
    packet_id: int
    current_lap: int
    total_laps: int
    best_lap_time_ms: int
    last_lap_time_ms: int
    time_of_day_ms: int
    race_start_pos: int
    total_cars: int
    min_alert_rpm: int
    max_alert_rpm: int
    calc_max_speed: int
    flags: int
    bits: int
    throttle: int
    brake: int
    empty: int
    road_plane_x: float
    road_plane_y: float
    road_plane_z: float
    road_plane_dist: float
    wheel_fl_rps: float
    wheel_fr_rps: float
    wheel_rl_rps: float
    wheel_rr_rps: float
    tire_fl_radius: float
    tire_fr_radius: float
    tire_rl_radius: float
    tire_rr_radius: float
    tire_fl_sus_height: float
    tire_fr_sus_height: float
    tire_rl_sus_height: float
    tire_rr_sus_height: float
    unused1: int
    unused2: int
    unused3: int
    unused4: int
    unused5: int
    unused6: int
    unused7: int
    unused8: int
    clutch_pedal: float
    clutch_engagement: float
    trans_rpm: float
    trans_top_speed: float
    gear1: float
    gear2: float
    gear3: float
    gear4: float
    gear5: float
    gear6: float
    gear7: float
    gear8: float
    car_code: int



from __future__ import annotations

import struct

from salsa20 import Salsa20_xor


class PDEncyption:
    """
    PDEncyption class provides methods for decrypting ciphertext using Salsa20 stream cipher.
    Credit to https://github.com/Nenkai/PDTools
    """

    _DEFAULT_KEY = b"Simulator Interface Packet ver 0.0"
    _GT7_KEY = b"Simulator Interface Packet GT7 ver 0.0"
    _IV_MASK: int = 0xDEADBEAF

    def __init__(self, is_gt7):
        self.is_gt7 = is_gt7

    def decrypt(self, ciphertext: bytes) -> bytes:
        """
        Decrypts the provided ciphertext using Salsa20 stream cipher.

        Parameters:
        - ciphertext (bytes): The encrypted data to be decrypted.
        - is_gt7 (bool): Flag indicating whether to use the GT7 key. Default is True.

        Returns:
        bytes: The decrypted plaintext.
        """
        seed = struct.unpack("<I", ciphertext[0x40:0x44])[0]
        iv = seed ^ self._IV_MASK
        iv = struct.pack("<II", iv, seed)
        return Salsa20_xor(
            ciphertext, iv, self._GT7_KEY[:32] if self.is_gt7 else self._DEFAULT_KEY[:32]
        )



from __future__ import annotations

import asyncio
import logging
from enum import Enum
from typing import Callable, List

from gt_telem.turismo_client import TurismoClient


class GameState(Enum):
    NOT_RUNNING = 0,
    RUNNING = 1,
    IN_MENU = 2,
    AT_TRACK = 3,
    IN_RACE = 4,
    PAUSED = 5,
    END_RACE = 6


class TurismoGame:
    on_running: list[Callable] = []
    on_in_game_menu: list[Callable] = []
    on_at_track: list[Callable] = []
    on_in_race: list[Callable] = []
    on_paused: list[Callable] = []
    on_race_end: list[Callable] = []

    def __init__(self, tc: TurismoClient):
        self.game_state: GameState = GameState.NOT_RUNNING
        self.check_next = 0
        self.tod = tc.telemetry.time_of_day_ms if tc.telemetry else 0
        tc.register_callback(TurismoGame._state_tracker, [self])

    def _change_state(self, state: GameState):
        events = None
        match state:
            case GameState.RUNNING:
                events = self.on_running
            case GameState.IN_MENU:
                events = self.on_in_game_menu
            case GameState.AT_TRACK:
                events = self.on_at_track
            case GameState.IN_RACE:
                events = self.on_in_race
            case GameState.PAUSED:
                events = self.on_paused
            case GameState.END_RACE:
                events = self.on_race_end
            case _:
                events = []
        for event in events:
            event()
        self.game_state = state

    @staticmethod
    async def _state_tracker(t, context):
        """
        Try to keep track of what the game is doing. This is remarkably challenging
        and probably not reliable. Things like replays, licenses and non-standard race events
        aren't handled well.
        """
        self = context
        if self.game_state == GameState.NOT_RUNNING:
            if t is None:
                return
            self._change_state(GameState.RUNNING)
            logging.debug("Game is running.")
        last = self.game_state

        match self.game_state:
            case GameState.RUNNING:
                if t.cars_on_track:
                    if t.is_paused:
                        logging.debug("Race is paused.")
                        self._change_state(GameState.PAUSED)
                    if t.current_lap == -1:
                        logging.debug("End of a Race.")
                        self._change_state(GameState.END_RACE)
                    else:
                        logging.debug("In a Race.")
                        self._change_state(GameState.IN_RACE)
                else:
                    if t.current_lap > -1:
                        logging.debug("At a track")
                        self._change_state(GameState.AT_TRACK)
                    else:
                        logging.debug("In a menu.")
                        self._change_state(GameState.IN_MENU)

            case GameState.IN_MENU:
                if t.current_lap > -1:
                    logging.debug("At a track")
                    self._change_state(GameState.AT_TRACK)

            case GameState.AT_TRACK:
                if t.cars_on_track:
                    logging.debug("In a Race.")
                    self._change_state(GameState.IN_RACE)
                elif t.current_lap == -1:
                    logging.debug("In a menu.")
                    self._change_state(GameState.IN_MENU)

            case GameState.IN_RACE:
                if t.is_paused:
                    logging.debug("Race is paused.")
                    self._change_state(GameState.PAUSED)
                elif t.current_lap == -1:
                    logging.debug("Race ended.")
                    self._change_state(GameState.END_RACE)

            case GameState.PAUSED:
                if not t.is_paused:
                    # The only way to really tell if they're resuming or exiting
                    # is to compare two frames. Engine RPM is a good candidate to
                    # test if the simulation is still running because even when stopped
                    # it fluctuates. This means losing a few frames but if that's a
                    # problem, don't pause the game :)
                    if self.check_next == 0:
                        self.check_next = 2
                    else:
                        if self.check_next > 1:
                            self.engine_rpm = t.engine_rpm
                            self.check_next -= 1
                            return
                        else:
                            self.check_next = 0
                            if t.engine_rpm != self.engine_rpm:
                                logging.debug("Resuming Race")
                                self._change_state(GameState.IN_RACE)
                            else:
                                logging.debug("Quit Race")
                                self._change_state(GameState.END_RACE)

            case GameState.END_RACE:
                if not t.is_loading:
                    logging.debug("At a track")
                    self._change_state(GameState.AT_TRACK)

            case _:
                logging.debug(f"Unhandled state: {self.game_state}")



from __future__ import annotations

import asyncio
import copy
import logging
import socket
import threading
from collections import deque
from time import sleep

from gt_telem.errors.playstation_errors import (PlayStationNotFoundError,
                                                PlayStatonOnStandbyError)
from gt_telem.models.helpers import SpanReader
from gt_telem.models.telemetry import Telemetry
from gt_telem.net.crypto import PDEncyption
from gt_telem.net.device_discover import get_ps_ip_type


class TurismoClient:
    RECEIVE_PORT = 33339
    BIND_PORT = 33340

    def __init__(self, is_gt7: bool=True, ps_ip: str=None):
        """
        Initialize TurismoClient.

        Parameters:
            - is_gt7 (bool): Flag indicating whether it's Gran Turismo 7. Default is True.
            - ps_ip (str): PlayStation IP address. If None, it will be discovered.
            - cancellation_token (asyncio.Event): Set token to shut down threads and return from run()
        """
        self._cancellation_token = None
        ip, ps = get_ps_ip_type()
        ip = ip or ps_ip
        if not ip:
            raise PlayStationNotFoundError()
        if ps and "STANDBY" in ps:
            raise PlayStatonOnStandbyError(ip)

        logging.info(f"Using the {ps} at {ip}")
        self.ip_addr: str = ip

        if is_gt7:
            self.RECEIVE_PORT += 400
            self.BIND_PORT += 400
        self._crypto: PDEncyption = PDEncyption(is_gt7)

        self._telem_lock: threading.Lock = threading.Lock()
        self._telem: Telemetry = None

        self._telem_update_callbacks = {}
        self._telem_callback_queue = asyncio.LifoQueue(maxsize=1)
        self._processing_callbacks = False

    @property
    def telemetry(self) -> Telemetry:
        """
        Get a copy of the telemetry data.

        Returns:
            Telemetry: A copy of the telemetry data.
        """
        if not self._telem:
            return None
        with self._telem_lock:
            cpy: Telemetry = copy.deepcopy(self._telem)
        return cpy

    @telemetry.setter
    def telemetry(self, value: Telemetry) -> None:
        """
        Set the telemetry data and call any registered callbacks.

        Parameters:
            - value (Telemetry): Telemetry data to set.
        """
        with self._telem_lock:
            self._telem = value

        try:
            self._telem_callback_queue.put_nowait(value)
        except asyncio.QueueFull:
            self._telem_callback_queue.get_nowait()
            self._telem_callback_queue.put_nowait(value)
        if not self._processing_callbacks:
            asyncio.create_task(self._process_telemetry_callbacks())

    def register_callback(self, callback, args=None):
        """
        Register an awaitable callback to be invoked when new telemetry is received.

        The telemetry object is sent as the first parameter, and additional
        args can be passed if specified.

        Callbacks are executed off the main thread, potentially compromising
        state integrity (e.g., using `self.` within your callback won't work).
        To work around this limitation, declare your callback as a @staticmethod,
        pass the class instance (self) as an argument, and receive the context of
        the class in your parameters (after telemetry, which is the first).

        .. code-block:: python

            def __init__(self, tc: TurismoClient):
                tc.add_callback(MyClass.parse_telem, [self])

            @staticmethod
            async def parse_telem(t: Telemetry, context: MyClass):
                self = context


        Additionally, note that the game sends telemetry at the same frequency as
        the frame rate (~60/s). If your callback takes too long to process and exit,
        subsequent callbacks will not be invoked until it returns.
        """
        self._telem_update_callbacks[callback] = args

    def deregister_callback(self, callback):
        """
        Deregister a callback.

        Parameters:
            - callback: Callback to deregister.
        """
        self._telem_update_callbacks.pop(callback)

    def run(self, cancellation_token: asyncio.Event=None) -> None:
        """
        Start the telemetry client and run the event loop.

        Parameters:
            - cancellation_token (asyncio.Event): Set token to shut down threads and return from run()
        """
        self._cancellation_token = cancellation_token or asyncio.Event()
        loop = asyncio.get_event_loop()
        heartbeat_task = loop.create_task(self._send_heartbeat())
        listen_task = loop.create_task(self._listen(loop))

        # Run the tasks in the event loop
        try:
            loop.run_until_complete(asyncio.gather(heartbeat_task, listen_task))
        except KeyboardInterrupt:
            self._cancellation_token.set()
        finally:
            # Clean up any resources here if needed
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    async def _send_heartbeat(self) -> None:
        """
        Send heartbeat messages at regular intervals to keep the telemetry stream alive.
        """
        logging.info("Starting telemetry heartbeat.")
        msg: bytes = b"A"
        while not self._cancellation_token.is_set():
            udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            udp_socket.sendto(msg, (self.ip_addr, self.RECEIVE_PORT))
            udp_socket.close()
            await asyncio.sleep(10)

    async def _listen(self, loop: asyncio.AbstractEventLoop) -> None:
        """
        Listen for incoming telemetry data.

        Parameters:
            - loop: The asyncio event loop.
        """
        logging.info(f"Listening for data on {self.ip_addr}:{self.BIND_PORT}")

        class MyDatagramProtocol(asyncio.DatagramProtocol):
            def __init__(self, client):
                self.client = client

            def datagram_received(self, data, addr):
                self.client._handle_data(data)

        udp_socket, _ = await loop.create_datagram_endpoint(
            lambda: MyDatagramProtocol(self),
            local_addr=("0.0.0.0", self.BIND_PORT)
        )
        await self._cancellation_token.wait()
        udp_socket.close()

    async def _process_telemetry_callbacks(self):
        """
        Process telemetry callbacks.
        """
        self._processing_callbacks = True
        while True:
            try:
                # Wait for the next telemetry update callback
                telemetry_value = await self._telem_callback_queue.get()

                # Call the user-provided callback
                for cb, args in self._telem_update_callbacks.items():
                    if args:
                        await cb(telemetry_value, *args)
                    else:
                        await cb(telemetry_value)

                # Optionally introduce a delay here if needed
                await asyncio.sleep(1 / 60)  # 60 Hz update rate

            except asyncio.CancelledError:
                # The task is cancelled when the event loop is stopped
                break
            except Exception as e:
                # Handle exceptions during callback processing
                logging.error(f"Error processing telemetry {cb}: {e}")

        self._processing_callbacks = False

    def _handle_data(self, data: bytes) -> None:
        """
        Handle incoming telemetry data.

        Parameters:
            - data: Raw telemetry data.
        """
        try:
            message: bytes = self._crypto.decrypt(data)
        except Exception as e:
            logging.debug(f"Failed to decrypt. Error: {e}. Wrong system?")
            return
        # First 4 bytes are header and indicate which system this is
        try:
            header: str = message[:4].decode("ascii")
        except Exception as e:
            logging.debug(f"Not sure what this is \n{message[:4]}. Error: {e}")
            return
        message: bytes = message[4:]
        if not header in ["0S7G", "G6S0"]:
            # bad data
            logging.debug(f"Not sure what this is \n{header}")
            return
        if header == "0S7G":
            sr: SpanReader = SpanReader(message, "little")
        else:
            sr: SpanReader = SpanReader(message, "big")
        self._parse_telemetry(sr)

    def _parse_telemetry(self, sr: SpanReader) -> None:
        """
        Parse telemetry data from SpanReader and update the telemetry property.

        Parameters:
            - sr: SpanReader containing telemetry data.
        """
        self.telemetry = Telemetry(
            position_x=sr.read_single(),
            position_y=sr.read_single(),
            position_z=sr.read_single(),
            velocity_x=sr.read_single(),
            velocity_y=sr.read_single(),
            velocity_z=sr.read_single(),
            rotation_x=sr.read_single(),
            rotation_y=sr.read_single(),
            rotation_z=sr.read_single(),
            orientation=sr.read_single(),
            ang_vel_x=sr.read_single(),
            ang_vel_y=sr.read_single(),
            ang_vel_z=sr.read_single(),
            body_height=sr.read_single(),
            engine_rpm=sr.read_single(),
            iv=sr.read_single(),
            fuel_level=sr.read_single(),
            fuel_capacity=sr.read_single(),
            speed_mps=sr.read_single(),
            boost_pressure=sr.read_single(),
            oil_pressure=sr.read_single(),
            water_temp=sr.read_single(),
            oil_temp=sr.read_single(),
            tire_fl_temp=sr.read_single(),
            tire_fr_temp=sr.read_single(),
            tire_rl_temp=sr.read_single(),
            tire_rr_temp=sr.read_single(),
            packet_id=sr.read_int32(),
            current_lap=sr.read_int16(),
            total_laps=sr.read_int16(),
            best_lap_time_ms=sr.read_int32(),
            last_lap_time_ms=sr.read_int32(),
            time_of_day_ms=sr.read_int32(),
            race_start_pos=sr.read_int16(),
            total_cars=sr.read_int16(),
            min_alert_rpm=sr.read_int16(),
            max_alert_rpm=sr.read_int16(),
            calc_max_speed=sr.read_int16(),
            flags=sr.read_int16(),
            bits=sr.read_byte(),
            throttle=sr.read_byte(),
            brake=sr.read_byte(),
            empty=sr.read_byte(),
            road_plane_x=sr.read_single(),
            road_plane_y=sr.read_single(),
            road_plane_z=sr.read_single(),
            road_plane_dist=sr.read_single(),
            wheel_fl_rps=sr.read_single(),
            wheel_fr_rps=sr.read_single(),
            wheel_rl_rps=sr.read_single(),
            wheel_rr_rps=sr.read_single(),
            tire_fl_radius=sr.read_single(),
            tire_fr_radius=sr.read_single(),
            tire_rl_radius=sr.read_single(),
            tire_rr_radius=sr.read_single(),
            tire_fl_sus_height=sr.read_single(),
            tire_fr_sus_height=sr.read_single(),
            tire_rl_sus_height=sr.read_single(),
            tire_rr_sus_height=sr.read_single(),
            unused1=sr.read_int32(),
            unused2=sr.read_int32(),
            unused3=sr.read_int32(),
            unused4=sr.read_int32(),
            unused5=sr.read_int32(),
            unused6=sr.read_int32(),
            unused7=sr.read_int32(),
            unused8=sr.read_int32(),
            clutch_pedal=sr.read_single(),
            clutch_engagement=sr.read_single(),
            trans_rpm=sr.read_single(),
            trans_top_speed=sr.read_single(),
            gear1=sr.read_single(),
            gear2=sr.read_single(),
            gear3=sr.read_single(),
            gear4=sr.read_single(),
            gear5=sr.read_single(),
            gear6=sr.read_single(),
            gear7=sr.read_single(),
            gear8=sr.read_single(),
            car_code=sr.read_int32(),
        )



from __future__ import annotations

import socket
from typing import Optional, Tuple


def _get_host_type(buffer: str) -> str:
    for line in buffer.split("\n"):
        if line.startswith("HTTP"):
            code = int(line.split()[1])
        elif line:
            key, value = map(str.strip, line.split(":", 1))
            if key == "host-type":
                ps_type = value
                break
    return ps_type + " STANDBY" if code == 620 else ps_type


def get_ps_ip_type() -> tuple[str | None, str | None]:
    """
    Discovers the PlayStation IP address and host type using device discovery protocol.

    Returns:
    Tuple[Optional[str], Optional[str]]: A tuple containing the PlayStation IP address and host type.
    """
    skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    skt.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    skt.settimeout(1)

    query = b"SRCH * HTTP/1.1\ndevice-discovery-protocol-version:00030010"

    skt.sendto(query, ("<broadcast>", 9302))
    try:
        packet, addr = skt.recvfrom(1024)
    except socket.timeout:
        return None, None

    ps_type = _get_host_type(packet.decode("utf-8"))
    host_ip = addr[0]

    return host_ip, ps_type



from __future__ import annotations

from .helpers import SpanReader, format_time, format_time_of_day
from .models import Vector3D, WheelMetric
from .telemetry import Telemetry
from .telemetry_packet import TelemetryPacket



from __future__ import annotations

from datetime import datetime

from gt_telem.models.helpers import format_time, format_time_of_day
from gt_telem.models.models import Vector3D, WheelMetric
from gt_telem.models.telemetry_packet import TelemetryPacket


class Telemetry(TelemetryPacket):
    """
    Telemetry data from Gran Turismo

    Attributes:
    - position_x: float - X-coordinate of the position.
    - position_y: float - Y-coordinate of the position.
    - position_z: float - Z-coordinate of the position.
    - velocity_x: float - X-component of velocity.
    - velocity_y: float - Y-component of velocity.
    - velocity_z: float - Z-component of velocity.
    - rotation_x: float - X-component of rotation.
    - rotation_y: float - Y-component of rotation.
    - rotation_z: float - Z-component of rotation.
    - orientation: float - Orientation.
    - ang_vel_x: float - X-component of angular velocity.
    - ang_vel_y: float - Y-component of angular velocity.
    - ang_vel_z: float - Z-component of angular velocity.
    - body_height: float - Height of the body.
    - engine_rpm: float - Engine RPM.
    - iv: float - IV, used for encryption.
    - fuel_level: float - Fuel level.
    - fuel_capacity: float - Fuel capacity.
    - speed_mps: float - Speed in meters per second.
    - boost_pressure: float - Boost pressure.
    - oil_pressure: float - Oil pressure.
    - water_temp: float - Water temperature.
    - oil_temp: float - Oil temperature.
    - tire_fl_temp: float - Front-left tire temperature.
    - tire_fr_temp: float - Front-right tire temperature.
    - tire_rl_temp: float - Rear-left tire temperature.
    - tire_rr_temp: float - Rear-right tire temperature.
    - packet_id: int - Packet ID.
    - current_lap: int - Current lap.
    - total_laps: int - Total laps.
    - best_lap_time_ms: int - Best lap time in milliseconds.
    - last_lap_time_ms: int - Last lap time in milliseconds.
    - time_of_day_ms: int - Time of day in milliseconds.
    - race_start_pos: int - Race start position.
    - total_cars: int - Total number of cars.
    - min_alert_rpm: int - Minimum alert RPM.
    - max_alert_rpm: int - Maximum alert RPM.
    - calc_max_speed: int - Calculated maximum speed.
    - flags: int - byte that contains current/suggested gear.
    - bits: int - Collection of booleans - see properties.
    - throttle: int - Throttle.
    - brake: int - Brake.
    - empty: int - Unused.
    - road_plane_x: float - X-coordinate of the road plane.
    - road_plane_y: float - Y-coordinate of the road plane.
    - road_plane_z: float - Z-coordinate of the road plane.
    - road_plane_dist: float - Distance of the road plane. Not sure what this is.
    - wheel_fl_rps: float - Front-left wheel revolutions per second.
    - wheel_fr_rps: float - Front-right wheel revolutions per second.
    - wheel_rl_rps: float - Rear-left wheel revolutions per second.
    - wheel_rr_rps: float - Rear-right wheel revolutions per second.
    - tire_fl_radius: float - Front-left tire radius.
    - tire_fr_radius: float - Front-right tire radius.
    - tire_rl_radius: float - Rear-left tire radius.
    - tire_rr_radius: float - Rear-right tire radius.
    - tire_fl_sus_height: float - Front-left tire suspension height.
    - tire_fr_sus_height: float - Front-right tire suspension height.
    - tire_rl_sus_height: float - Rear-left tire suspension height.
    - tire_rr_sus_height: float - Rear-right tire suspension height.
    - unused1: int - Unused variable 1.
    - unused2: int - Unused variable 2.
    - unused3: int - Unused variable 3.
    - unused4: int - Unused variable 4.
    - unused5: int - Unused variable 5.
    - unused6: int - Unused variable 6.
    - unused7: int - Unused variable 7.
    - unused8: int - Unused variable 8.
    - clutch_pedal: float - Clutch pedal position.
    - clutch_engagement: float - Clutch engagement.
    - trans_rpm: float - Transmission RPM.
    - trans_top_speed: float - Transmission top speed.
    - gear1: float - Gear 1.
    - gear2: float - Gear 2.
    - gear3: float - Gear 3.
    - gear4: float - Gear 4.
    - gear5: float - Gear 5.
    - gear6: float - Gear 6.
    - gear7: float - Gear 7.
    - gear8: float - Gear 8.
    - car_code: int - Car code - on vehicles with more than 8 gears, this is corrupted.

    Properties:
    - position: Get the position as a Vector3D.
    - velocity: Get the velocity as a Vector3D.
    - rotation: Get the rotation as a Vector3D.
    - angular_velocity: Get the angular velocity as a Vector3D.
    - road_plane: Get the road plane coordinates as a Vector3D.
    - tire_temp: Get tire temperatures as a WheelMetric.
    - wheel_rps: Get wheel revolutions per second as a WheelMetric.
    - tire_radius: Get tire radii as a WheelMetric.
    - suspension_height: Get suspension heights as a WheelMetric.
    - current_gear: Get the current gear.
    - suggested_gear: Get the suggested gear.
    - speed_kph: Get the speed in kilometers per hour.
    - speed_mph: Get the speed in miles per hour.
    - cars_on_track: Check if there are cars on the track.
    - is_paused: Check if the simulation is paused.
    - is_loading: Check if the simulation is loading.
    - in_gear: Check if the vehicle is in gear.
    - has_turbo: Check if the vehicle has a turbo.
    - rev_limit: Check if the vehicle is at the rev limit.
    - hand_brake_active: Check if the hand brake is active.
    - lights_active: Check if the lights are active.
    - high_beams: Check if the high beams are active.
    - low_beams: Check if the low beams are active.
    - asm_active: Check if the ASM (Active Stability Management) is active.
    - tcs_active: Check if the TCS (Traction Control System) is active.
    - unknown_bool_1: Usually true. Purpose unknown.
    - clutch_out: Check if the clutch is out.
    - unknown_bool_3: Usually true. Purpose unknown.
    - best_lap_time: Get the formatted best lap time.
    - last_lap_time: Get the formatted last lap time.
    - time_of_day: Get the formatted time of day.

    Methods
    - as_dict: Get the state of the object in a dictionary format.
    """

    def __post_init__(self):
        self.time = datetime.now()

    @property
    def position(self) -> Vector3D:
        """
        Get the position as a Vector3D.
        """
        return Vector3D(self.position_x, self.position_y, self.position_z)

    @property
    def velocity(self) -> Vector3D:
        """
        Get the velocity as a Vector3D.
        """
        return Vector3D(self.velocity_x, self.velocity_y, self.velocity_z)

    @property
    def rotation(self) -> Vector3D:
        """
        Get the rotation as a Vector3D.
        """
        return Vector3D(self.rotation_x, self.rotation_y, self.rotation_z)

    @property
    def angular_velocity(self) -> Vector3D:
        """
        Get the angular velocity as a Vector3D.
        """
        return Vector3D(self.ang_vel_x, self.ang_vel_y, self.ang_vel_z)

    @property
    def road_plane(self) -> Vector3D:
        """
        Get the road plane coordinates as a Vector3D.
        """
        return Vector3D(self.road_plane_x, self.road_plane_y, self.road_plane_z)

    @property
    def tire_temp(self) -> WheelMetric:
        """
        Get tire temperatures as a WheelMetric.
        """
        return WheelMetric(
            self.tire_fl_temp, self.tire_fr_temp, self.tire_rl_temp, self.tire_rr_temp
        )

    @property
    def wheel_rps(self) -> WheelMetric:
        """
        Get wheel revolutions per second as a WheelMetric.
        """
        return WheelMetric(
            self.wheel_fl_rps, self.wheel_fr_rps, self.wheel_rl_rps, self.wheel_rr_rps
        )

    @property
    def tire_radius(self) -> WheelMetric:
        """
        Get tire radii as a WheelMetric.
        """
        return WheelMetric(
            self.tire_fl_radius,
            self.tire_fr_radius,
            self.tire_rl_radius,
            self.tire_rr_radius,
        )

    @property
    def suspension_height(self) -> WheelMetric:
        """
        Get suspension heights as a WheelMetric.
        """
        return WheelMetric(
            self.tire_fl_sus_height,
            self.tire_fr_sus_height,
            self.tire_rl_sus_height,
            self.tire_rr_sus_height,
        )

    @property
    def current_gear(self) -> int:
        """
        Get the current gear.
        """
        return self.bits & 0b1111

    @property
    def suggested_gear(self) -> int:
        """
        Get the suggested gear.
        """
        return self.bits >> 4

    @property
    def speed_kph(self) -> float:
        """
        Get the speed in kilometers per hour.
        """
        return self.speed_mps * 3.6

    @property
    def speed_mph(self) -> float:
        """
        Get the speed in miles per hour.
        """
        return self.speed_mps * 2.23694

    @property
    def cars_on_track(self) -> bool:
        """
        Check if there are cars on the track.
        """
        return bool(1 & self.flags)

    @property
    def is_paused(self) -> bool:
        """
        Check if the simulation is paused.
        """
        return bool(2 & self.flags)

    @property
    def is_loading(self) -> bool:
        """
        Check if the simulation is loading.
        """
        return bool(3 & self.flags)

    @property
    def in_gear(self) -> bool:
        """
        Check if the vehicle is in gear.
        """
        return bool(4 & self.flags)

    @property
    def has_turbo(self) -> bool:
        """
        Check if the vehicle has a turbo.
        """
        return bool(5 & self.flags)

    @property
    def rev_limit(self) -> bool:
        """
        Check if the vehicle is at the rev limit.
        """
        return bool(6 & self.flags)

    @property
    def hand_brake_active(self) -> bool:
        """
        Check if the hand brake is active.
        """
        return bool(7 & self.flags)

    @property
    def lights_active(self) -> bool:
        """
        Check if the lights are active.
        """
        return bool(8 & self.flags)

    @property
    def high_beams(self) -> bool:
        """
        Check if the high beams are active.
        """
        return bool(9 & self.flags)

    @property
    def low_beams(self) -> bool:
        """
        Check if the low beams are active.
        """
        return bool(10 & self.flags)

    @property
    def asm_active(self) -> bool:
        """
        Check if the ASM (Active Stability Management) is active.
        """
        return bool(11 & self.flags)

    @property
    def tcs_active(self) -> bool:
        """
        Check if the TCS (Traction Control System) is active.
        """
        return bool(12 & self.flags)

    @property
    def unknown_bool_1(self) -> bool:
        """
        Get the value of an unknown boolean flag.
        """
        return bool(13 & self.flags)

    @property
    def clutch_out(self) -> bool:
        """
        Check if the clutch is out.
        """
        return bool(14 & self.flags)

    @property
    def unknown_bool_3(self) -> bool:
        """
        Get the value of another unknown boolean flag.
        """
        return bool(15 & self.flags)

    @property
    def best_lap_time(self) -> str:
        """
        Get the formatted best lap time.
        """
        if self.best_lap_time_ms == -1:
            return None
        return format_time(self.best_lap_time_ms)

    @property
    def last_lap_time(self) -> str:
        """
        Get the formatted last lap time.
        """
        if self.last_lap_time_ms == -1:
            return None
        return format_time(self.last_lap_time_ms)

    @property
    def time_of_day(self) -> str:
        """
        Get the formatted time of day.
        """
        if self.time_of_day_ms == -1:
            return None
        return format_time_of_day(self.time_of_day_ms)

    @property
    def as_dict(self):
        """
        Returns a dictionary containing the state of the object.
        """
        remove_keys = [
            x
            for x in self.__dict__.keys()
            if any(
                ignore in x
                for ignore in [
                    "_x",
                    "_y",
                    "_z",
                    "flags",
                    "bits",
                    "empty",
                    "unused",
                    "_fl",
                    "_fr",
                    "_rl",
                    "_rr",
                ]
            )
        ]

        added = {
            "position": self.position,
            "velocity": self.velocity,
            "rotation": self.rotation,
            "angular_velocity": self.angular_velocity,
            "road_plane": self.road_plane,
            "current_gear": self.current_gear,
            "suggested_gear": self.suggested_gear,
            "speed_kph": self.speed_kph,
            "speed_mph": self.speed_mph,
            "cars_on_track": self.cars_on_track,
            "is_paused": self.is_paused,
            "is_loading": self.is_loading,
            "in_gear": self.in_gear,
            "has_turbo": self.has_turbo,
            "rev_limit": self.rev_limit,
            "hand_brake_active": self.hand_brake_active,
            "lights_active": self.lights_active,
            "high_beams": self.high_beams,
            "low_beams": self.low_beams,
            "asm_active": self.asm_active,
            "tcs_active": self.tcs_active,
            "unknown_bool_1": self.unknown_bool_1,
            "clutch_out": self.clutch_out,
            "unknown_bool_3": self.unknown_bool_3,
            "best_lap_time": self.best_lap_time,
            "last_lap_time": self.last_lap_time,
            "time_of_day": self.time_of_day,
        }

        result = dict(self.__dict__, **added)
        for remove_key in remove_keys:
            result.pop(remove_key, None)

        return result



from __future__ import annotations

import struct


def format_time(milliseconds):
    """
    Format milliseconds into a time string (MM:SS.sss).

    Parameters:
        - milliseconds (int): Time in milliseconds.

    Returns:
        str: Formatted time string.
    """
    milliseconds = max(0, int(milliseconds))
    minutes, milliseconds = divmod(milliseconds, 60000)
    seconds, milliseconds = divmod(milliseconds, 1000)
    return f"{minutes:02}:{seconds:02}.{milliseconds:03}"


def format_time_of_day(milliseconds, use_24hr=False):
    """
    Format milliseconds into a time of day string.

    Parameters:
        - milliseconds (int): Time in milliseconds.
        - use_24hr (bool): Flag indicating whether to use 24-hour format. Default is False.

    Returns:
        str: Formatted time of day string.
    """
    milliseconds = max(0, int(milliseconds))
    hours, milliseconds = divmod(milliseconds, 3600000)
    minutes, milliseconds = divmod(milliseconds, 60000)
    seconds, milliseconds = divmod(milliseconds, 1000)

    if use_24hr:
        am_pm = "AM" if hours < 12 else "PM"
        hours = hours % 12 or 12
        return f"{hours:02}:{minutes:02}:{seconds:02} {am_pm}"
    else:
        return f"{hours:02}:{minutes:02}:{seconds:02}"


class SpanReader:
    """
    Utility class for reading binary data in a structured manner.
    """

    def __init__(self, data, byte_order="little"):
        """
        Initialize the SpanReader.

        Parameters:
            - data: Binary data to read.
            - byte_order (str): Byte order for interpreting binary data, 'little' or 'big'.
        """
        self.view = memoryview(data)
        self.byte_order = "<" if byte_order == "little" else ">"
        self.position = 0

    def read_int32(self):
        """
        Read a 32-bit signed integer from the binary data.

        Returns:
            int: The read integer value.
        """
        format_string = f"{self.byte_order}i"
        value = struct.unpack_from(format_string, self.view, self.position)[0]
        self.position += 4
        return value

    def read_int16(self):
        """
        Read a 16-bit signed integer from the binary data.

        Returns:
            int: The read integer value.
        """
        format_string = f"{self.byte_order}h"
        value = struct.unpack_from(format_string, self.view, self.position)[0]
        self.position += 2
        return value

    def read_single(self):
        """
        Read a 32-bit floating-point number from the binary data.

        Returns:
            float: The read floating-point value.
        """
        format_string = f"{self.byte_order}f"
        value = struct.unpack_from(format_string, self.view, self.position)[0]
        self.position += 4
        return value

    def read_byte(self):
        """
        Read a byte from the binary data.

        Returns:
            int: The read byte value.
        """
        value = struct.unpack_from("B", self.view, self.position)[0]
        self.position += 1
        return value

    def read_bytes(self, length):
        """
        Read a specified number of bytes from the binary data.

        Parameters:
            - length (int): Number of bytes to read.

        Returns:
            bytes: The read bytes.
        """
        value = self.view[self.position : self.position + length].tobytes()
        self.position += length
        return value



from __future__ import annotations

import unittest
from datetime import datetime

from gt_telem.models import Telemetry, Vector3D


class TestTelemetry(unittest.TestCase):
    def setUp(self):
        self.telemetry = Telemetry(
            position_x=1.0,
            position_y=2.0,
            position_z=3.0,
            velocity_x=4.0,
            velocity_y=5.0,
            velocity_z=6.0,
            rotation_x=7.0,
            rotation_y=8.0,
            rotation_z=9.0,
            orientation=10.0,
            ang_vel_x=11.0,
            ang_vel_y=12.0,
            ang_vel_z=13.0,
            body_height=14.0,
            engine_rpm=15.0,
            iv=16.0,
            fuel_level=17.0,
            fuel_capacity=18.0,
            speed_mps=19.0,
            boost_pressure=20.0,
            oil_pressure=21.0,
            water_temp=22.0,
            oil_temp=23.0,
            tire_fl_temp=24.0,
            tire_fr_temp=25.0,
            tire_rl_temp=26.0,
            tire_rr_temp=27.0,
            packet_id=28,
            current_lap=29,
            total_laps=30,
            best_lap_time_ms=31,
            last_lap_time_ms=32,
            time_of_day_ms=33,
            race_start_pos=34,
            total_cars=35,
            min_alert_rpm=36,
            max_alert_rpm=37,
            calc_max_speed=38,
            flags=39,
            bits=40,
            throttle=41,
            brake=42,
            empty=43,
            road_plane_x=44.0,
            road_plane_y=45.0,
            road_plane_z=46.0,
            road_plane_dist=47.0,
            wheel_fl_rps=48.0,
            wheel_fr_rps=49.0,
            wheel_rl_rps=50.0,
            wheel_rr_rps=51.0,
            tire_fl_radius=52.0,
            tire_fr_radius=53.0,
            tire_rl_radius=54.0,
            tire_rr_radius=55.0,
            tire_fl_sus_height=56.0,
            tire_fr_sus_height=57.0,
            tire_rl_sus_height=58.0,
            tire_rr_sus_height=59.0,
            unused1=60,
            unused2=61,
            unused3=62,
            unused4=63,
            unused5=64,
            unused6=65,
            unused7=66,
            unused8=67,
            clutch_pedal=68.0,
            clutch_engagement=69.0,
            trans_rpm=70.0,
            trans_top_speed=71.0,
            gear1=72.0,
            gear2=73.0,
            gear3=74.0,
            gear4=75.0,
            gear5=76.0,
            gear6=77.0,
            gear7=78.0,
            gear8=79.0,
            car_code=80,
        )

    def test_position_property(self):
        self.assertEqual(self.telemetry.position, Vector3D(1.0, 2.0, 3.0))

    def test_velocity_property(self):
        self.assertEqual(self.telemetry.velocity, Vector3D(4.0, 5.0, 6.0))

    def test_time_formatting(self):
        pass

    def tearDown(self):
        # Clean up if needed
        pass


if __name__ == "__main__":
    unittest.main()
